local revision = tonumber(string.sub("$Revision: 85 $", 12, -3))
local rss = _G.RShamanStat
if rss.revision < revision then rss.revision = revision end

-- matches an AA with the spell that triggered it, based on closest match to the expected fraction
function rss.MatchAA(amount)
	local mindiff = amount; 
	local sMin = nil;
	local vMin = nil;
	for s, v in pairs(rss.playerState.lastCast) do
		if v.critical or rss.playerState.stats.aaSetBonusT15 then
			local fraction = v.critical and rss.SPELLS[rss.ANCESTRAL_AWAKENING].copy or rss.SET_BONUS_T15_FRACTION;
			if math.abs(v.amount*fraction - amount) < mindiff then
				mindiff = math.abs(v.amount*fraction - amount);
				vMin = v;
				sMin = s;
			end
		end
	end
	return sMin, vMin;
end

function rss.CalcCopy(record, amount, incrementalSpGain, masteryGain, incrementalMasteryGain, critical)
	if critical then
		record.rawHealing[1] = record.rawHealing[1] + amount;
	else
		record.rawHealing[2] = record.rawHealing[2] + amount;
	end
	record.incrementalSp = record.incrementalSp + incrementalSpGain;
	record.masteryTotal = record.masteryTotal + masteryGain;
	record.incrementalMastery = record.incrementalMastery + incrementalMasteryGain;
end

function rss.Leak(reason, healed)
	--print("RSS: Leakage detected");
	if not rss.fight.leakage[reason] then
		rss.fight.leakage[reason] = 0.0;
	end
	rss.fight.leakage[reason] = rss.fight.leakage[reason] + healed;
end

function rss.CombatLogEntry(...)
	local playerState = rss.playerState;
	local lastCast = playerState.lastCast;
	local fight = rss.fight;
	local timestamp, combatEvent, hideCaster, 
	      sourceGUID, sourceName, sourceFlags, sourceRaidFlags, 
	      destGUID, destName, destFlags, destRaidFlags = select(1, ...);
		  
	-- watch for the comings and goings of Spirit Link Totem
	if sourceGUID == playerState.GUID and combatEvent and strfind(combatEvent, 'SPELL_SUMMON') then
		local spellId, spellName, spellSchool = select(12, ...);
		if spellId == rss.SUMMON_SPIRIT_LINK then 
			playerState.spiritLinkGUID = destGUID;
		end
		return;
	end
	if combatEvent and strfind(combatEvent, 'UNIT_DESTROYED') and destGUID == playerState.spiritLinkGUID then
		playerState.spiritLinkGUID = nil;
		return;
    end

    -- handle power gains like resurgence
    if sourceGUID == playerState.GUID and combatEvent and combatEvent == "SPELL_ENERGIZE" then
        local spellId, spellName, spellSchool = select(12, ...)
        local amount, powerType = select(15, ...)
        if spellId == rss.RESURGENCE then
            fight.resurgence = fight.resurgence + amount
        elseif spellId == rss.PROC_WATER_SHIELD then
            fight.spells[rss.PROC_WATER_SHIELD] = (fight.spells[rss.PROC_WATER_SHIELD] or 0) + 1
        end
        return;
    end
	
	-- deal with the Flame Shock glyph
	if sourceGUID == playerState.GUID and combatEvent and strfind(combatEvent, '_DAMAGE') then
		local spellId = select(12,...);
		if spellId == rss.FLAME_SHOCK then
			playerState.lastFlameShock.combatEvent = combatEvent;
			playerState.lastFlameShock.critical = select(21,...);
		end
		return;
	end

	-- The remainder only looks at 
	-- 1) Heals
	-- 2) Cast by us or our SLT
	-- 3) On a target whose health we know
	local healthTable = rss.healthTable;
	local spiritLinkHeal = playerState.spiritLinkGUID and sourceGUID == playerState.spiritLinkGUID;
	if (sourceGUID == playerState.GUID or spiritLinkHeal) and combatEvent and strfind(combatEvent, '_HEAL') and healthTable[destGUID] then
		-- Check for boss encounter by looking at the target of anybody we heal
		if ( (not fight.isBoss) or (fight.encounterName == "Unknown") ) then
			rss.CheckBoss(destName);
		end

		-- Check for first action of combat
		if fight.startTime <= 0 then
            fight.startTime = timestamp;
		end
        fight.totalTime = timestamp - fight.startTime;
		
		-- We healed somebody; get the details
		local spellId, spellName, spellSchool = select(12, ...)
		local amount, overhealing, absorbed, critical = select(15, ...)
		if spellId == rss.FLAMES_OF_LIFE and playerState.lastFlameShock then
			critical = playerState.lastFlameShock.critical;
		elseif spellId == rss.GLYPH_OF_HEALING_WAVE and lastCast[rss.HEALING_WAVE] then
			critical = lastCast[rss.HEALING_WAVE].critical;
		end

		-- accumulate total effective throughput
		local healed = math.max(0.,amount - overhealing);
        fight.effectiveTotal = fight.effectiveTotal + healed;

		-- no stat effects on Spirit Link totem heals
		if spiritLinkHeal then 
			return;
        end

		-- Recognize consumables so they can be (mostly) ignored
		if tContains(rss.BANDAGES, spellId) then return; end
		local consumable = tContains(rss.CONSUMABLES, spellId);	
		
		-- Recognize totems
		local totem;
		if rss.SPELLS[spellId] and rss.SPELLS[spellId].totem then
			totem = true;
		end

		-- Recognize periodic healing
		local periodic;
		if spellId == rss.FLAMES_OF_LIFE then
			periodic = strfind(playerState.lastFlameShock.combatEvent, 'PERIODIC');
		else
			periodic = strfind(combatEvent, 'PERIODIC_HEAL');
		end

        -- count the number of times we healed with non-periodic spells
        if not periodic then
            local spellId_count = fight.spells[spellId] or 0
            spellId_count = spellId_count + 1
            fight.spells[spellId] = spellId_count
        end

		-- get the health missing from our target
		local deficit = math.max(healthTable[destGUID].maxHealth - healthTable[destGUID].currentHealth, 0);	
		
		-- deal with spell copies from Ascendance and Ancestral Guidance
		if spellId == rss.RESTORATIVE_MISTS or spellId == rss.ANCESTRAL_GUIDANCE then
			if playerState.ascendance and spellId == rss.RESTORATIVE_MISTS then
				local ascendance = playerState.ascendance;
				if (overhealing == 0) then
					ascendance.effectiveCopies = ascendance.effectiveCopies + healed;
				end
				ascendance.incrementalCritical[1] = rss.EvaluateCriticalRating(amount, deficit, true);
				ascendance.incrementalCritical[2] = rss.EvaluateCriticalRating(amount, deficit, false);
			elseif playerState.ancestralGuidance and spellId == rss.ANCESTRAL_GUIDANCE then
				local guidance = playerState.ancestralGuidance;
				if (overhealing == 0) then
					guidance.effectiveCopies = guidance.effectiveCopies + healed;	
				end
				guidance.incrementalCritical[1] = rss.EvaluateCriticalRating(amount, deficit, true);
				guidance.incrementalCritical[2] = rss.EvaluateCriticalRating(amount, deficit, false);
			else
				rss.Leak("Missing originals for copy", healed);  -- derp, we have no record of the spell(s) that this heal was copied from
			end
			return; -- nothing more to do for now; stats affect the spell that was copied, not the copy itself
		end
				
		-- match AA once for everything that follows
		local aaSpell = nil;
		local aaCast  = nil;
		local aaMultiplier = 0;
		if spellId == rss.ANCESTRAL_AWAKENING then
			aaSpell, aaCast = rss.MatchAA(amount);
			if aaCast == nil then
				rss.Leak("AA match not found", healed);
				return;
			end
			aaMultiplier = aaCast.critical and rss.SPELLS[spellId].copy or rss.SET_BONUS_T15_FRACTION;
		end
		
		-- Deal with Spellpower
		local mySpellPower = playerState.stats.healingBonus;
		local incrementalSp = 0.0;		
		if rss.SPELLS[spellId] and rss.SPELLS[spellId][combatEvent] then
			local c = rss.SPELLS[spellId][combatEvent];
			local baseAmt = (c[1]+c[2])/2 + mySpellPower*c[3];
			incrementalSp = (1.1*c[3]/baseAmt) * amount;   -- the 1.1 reflects our 10% spellpower buff
		elseif spellId == rss.ANCESTRAL_AWAKENING then
			incrementalSp = aaMultiplier * aaCast.incrementalSpellpower;
		elseif spellId == rss.GLYPH_OF_HEALING_WAVE then
			if lastCast[rss.HEALING_WAVE] then
				incrementalSp = (amount/lastCast[rss.HEALING_WAVE].amount) * lastCast[rss.HEALING_WAVE].incrementalSpellpower;
			else
				rss.Leak("Missing original HW for glyph heal", healed);
				return;
			end
		elseif spellId == rss.FLAMES_OF_LIFE and playerState.lastFlameShock and playerState.lastFlameShock.combatEvent then
			mySpellPower = playerState.stats.damageBonus[rss.SCHOOL_FIRE];
			local c = rss.SPELLS[spellId][playerState.lastFlameShock.combatEvent];
			local baseAmt = (c[1]+c[2])/2 + mySpellPower*c[3];
			incrementalSp = (1.1*c[3]/baseAmt) * amount;
		elseif not rss.SPELLS[spellId] and not consumable then
			if not rss.unknownSpells[spellId] then
				print("Unrecognized healing entry: "..spellName.." "..spellId.." "..combatEvent);
				rss.unknownSpells[spellId] = { name = spellName , count = 0 };
			end
			rss.unknownSpells[spellId].count = rss.unknownSpells[spellId].count + 1;
			consumable = true;  -- assume no stat benefits other than crit
		end
		if overhealing == 0 then
			fight.incrementalSpellpower = fight.incrementalSpellpower + incrementalSp;
		end
		
		-- Deal with Mastery
		local masteryGain = 0.0;
		local incrementalMasteryGain = 0.0;
		local hpratio = 0.0;
		-- find last known values of health and max health
		local unitHealthTable = healthTable[destGUID];
		if (not consumable) and rss.SPELLS[spellId] and (not rss.SPELLS[spellId].noMastery) and playerState.stats.masteryBonus > 0 then
			if spellId == rss.ANCESTRAL_AWAKENING then
				hpratio = aaCast.hpRatio;
				masteryGain = aaMultiplier * aaCast.masteryGain;
				incrementalMasteryGain = aaMultiplier * aaCast.incrementalMastery;
			elseif spellId == rss.GLYPH_OF_HEALING_WAVE then
				hpratio = lastCast[rss.HEALING_WAVE].hpRatio;			
				masteryGain = (amount/lastCast[rss.HEALING_WAVE].amount) * lastCast[rss.HEALING_WAVE].masteryGain;
				incrementalMasteryGain = (amount/lastCast[rss.HEALING_WAVE].amount) * lastCast[rss.HEALING_WAVE].incrementalMastery;
			else
				hpratio = 1 - (unitHealthTable.currentHealth/unitHealthTable.maxHealth);
				local m = playerState.stats.masteryBonus;
				masteryGain = amount * hpratio * m / (1 + hpratio * m);
				incrementalMasteryGain = (0.01/rss.RATING_TO_MASTERY) * masteryGain / m;     -- how much would +1 Mastery Rating increase this heal?
			end
			-- average the mastery efficiency factor over all relevant heals
			fight.weightedTotal = fight.weightedTotal + healed * hpratio
			fight.effectiveSubTotal = fight.effectiveSubTotal + healed
			-- accumulate gain from current Mastery
			fight.masteryTotal = fight.masteryTotal + math.max(0.0, masteryGain - overhealing);
			-- accumulate gain from +1 Mastery Rating
			if (overhealing == 0) then
				fight.incrementalMastery = fight.incrementalMastery + incrementalMasteryGain;       
			end	
		end
		
		-- Deal with Crit
		local aaAssumedDeficit;
		if spellId == rss.ANCESTRAL_AWAKENING then
			if aaCast.critical then
				local criticalIncrement, criticalContribution = rss.EvaluateAncestralAwakening(amount, deficit, true);
				fight.incrementalCritical = fight.incrementalCritical + criticalIncrement;
				fight.criticalTotal = fight.criticalTotal + criticalContribution;
			else
				local criticalIncrement = rss.CorrectAncestralAwakening(aaCast.amount, aaCast.assumedDeficit, amount, deficit);
				fight.incrementalCritical = fight.incrementalCritical + criticalIncrement;
			end
			return;
		elseif spellId == rss.GLYPH_OF_HEALING_WAVE then
			-- special calculation required, since the original heal AND the glyph (or both of them) can crit
			local lastHW = lastCast[rss.HEALING_WAVE];
			local criticalIncrement, criticalContribution = rss.GlyphOfHealingWaveCritical(lastHW.amount, amount, deficit, lastHW.critical);
			fight.incrementalCritical = fight.incrementalCritical + criticalIncrement;
			if critical or lastHW.critical then
				fight.criticalTotal = fight.criticalTotal + criticalContribution;
			end
			return;
		else	-- assume that anything that gets here is able to crit
			local criticalIncrement, criticalContribution = rss.EvaluateCriticalRating(amount, deficit, critical);
			if not critical and rss.SPELLS[spellId] and rss.SPELLS[spellId].aaTrigger then
				-- guess at the target an AA would hit if proc'ed
				aaAssumedDeficit = 0;
				for guid, t in pairs(healthTable) do
					local aaDeficit = math.max(0, t.maxHealth - t.currentHealth);
					if guid == destGUID then    -- account for the amount we *would* have healed the target...
						aaDeficit = math.max(0, aaDeficit - amount*playerState.stats.criticalMultiplier);
					end
					aaAssumedDeficit = math.max(aaAssumedDeficit, aaDeficit);
				end
				criticalIncrement = criticalIncrement + rss.EvaluateAncestralAwakening(amount, aaAssumedDeficit, false);
			end
			fight.incrementalCritical = fight.incrementalCritical + criticalIncrement;
			if critical then
				fight.criticalTotal = fight.criticalTotal + criticalContribution;
			end
		end
				
	-- record AA-eligible casts for later matching
		if rss.SPELLS[spellId] and ( rss.SPELLS[spellId].aaTrigger or spellId == rss.HEALING_WAVE ) and not periodic then
			local thisCast = { };
			thisCast.amount = amount;
			thisCast.hpRatio = hpratio;
			thisCast.spellPower = mySpellPower;
			thisCast.masteryGain = masteryGain;
			thisCast.incrementalMastery = incrementalMasteryGain;
			thisCast.incrementalSpellpower = incrementalSp;
			thisCast.critical = critical and true or false;
			thisCast.assumedDeficit = aaAssumedDeficit;
			lastCast[spellId] = thisCast;
		end
		
	-- store info for active copy buffs
		if (playerState.ascendanceAura or playerState.guidanceAura) and not totem then
			if playerState.ascendance then
				rss.CalcCopy(playerState.ascendance, amount, incrementalSp, masteryGain, incrementalMasteryGain, critical);
			end
			if playerState.ancestralGuidance and (not periodic) and (consumable or not rss.SPELLS[spellId] or not rss.SPELLS[spellId].noAG) then
				rss.CalcCopy(playerState.ancestralGuidance, amount, incrementalSp, masteryGain, incrementalMasteryGain, critical)
			end
		end
	end
end
